;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - properties-func.ss                                             #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Lozi Jean-Pierre, 2004 - mailto:jean-pierre@lozi.org             #
;; # Copyright (C) Peter Ivanyi, 2007                                               #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################


(module mreddesigner-properties-func mzscheme
  (require (lib "class.ss")  
           (lib "mred.ss" "mred")
           (lib "string.ss")
           (lib "list.ss")
           (file "mreddesigner-widgets.ss")
  )
  
  ;; This dialog box is shown when the user tries to set an invalid value to a property.
  (define (show-invalid-dialog-box field property)
    (if (= (string-length field) 0)
      (set! field "<nothing>")
    )
    (message-box "Error" (string-append "The value '"
                                        field
                                        "' is not valid for the '"
                                        property
                                        "' property.") properties-frame '(ok stop)))

  (define (remove-list props lst)
    (cond
      ((null? lst) '())
      ((member (car lst) props) (remove-list props (cdr lst)))
      (else (cons (car lst) (remove-list props (cdr lst))))
    )
  )
  
  ;; Returns true iff the string is valid for the property
  ; it returns #f or a list where 
  ;  car:  the name of the property
  ;  cadr: the actual value of the property
  (define (validate-property prop str)
    (define (style-prop-make prop)
      (let*
        ((selected (send hierarchy-widget get-selected))
         (style    (send hierarchy-widget item-property-get selected 'style #f))
         (exist?   (member prop style))
        )
        (cond
          ((and (string-ci=? str "true") (not exist?))
           (list 'style (cons prop style)))
          ((and (string-ci=? str "false") exist?)
           (list 'style (remove prop style)))
          (else (list 'style style))
        )
      )
    )
    (define (style-no-prop-make no-prop)
      (let*
        ((selected (send hierarchy-widget get-selected))
         (style    (send hierarchy-widget item-property-get selected 'style #f))
         (exist?   (member no-prop style))
        )
        (cond
          ((and (string-ci=? str "true") exist?)
           (list 'style (remove no-prop style)))
          ((and (string-ci=? str "false") (not exist?))
           (list 'style (cons no-prop style)))
          (else (list 'style style))
        )
      )
    )
    
    (case prop
      ;; If the property is a label, str is always valid.
      ((label)
       (if (= (string-length str) 0)
         (let*
           ((selected (send hierarchy-widget get-selected))
            (type     (send hierarchy-widget item-property-get selected 'type #f))
           )
           ; special widgets for which the label cannot be #f
           (if (member type '(button message group-box-panel frame check-box))
             (list prop "")
             ; text-field, slider, radio-box, list-box, gauge, combo-field, choice, list-box
             (list prop #f)
           )
         )
         (list prop str)
       )
      )
      ;; Those properties are positive integers in [0,10000].
      ((width height min-width min-height)
       (if (and (integer? (string->number str)) (<= 0 (string->number str) 10000))
         (list prop (string->number str))
         #f
       )
      )
      ;; Those properties are positive integers in [0,1000].
      ((vert-margin horiz-margin border spacing) 
       (if (and (integer? (string->number str)) (<= 0 (string->number str) 1000))
         (list prop (string->number str))
         #f
       )
      )
      ;; Those properties are integers in [-10000,10000].
      ((x y min-value max-value) 
       (if (and (integer? (string->number str)) (<= -10000 (string->number str) 10000))
         (list prop (string->number str))
         #f
       )
      )
      ;; This property is an integer in [1,10000].
      ((range)
       (if (and (integer? (string->number str)) (<= 1 (string->number str) 10000))
         (list prop (string->number str))
         #f
       )
      )
      ;; These properties are booleans.
      ((enabled stretchable-width stretchable-height value)
       (cond
         ((string-ci=? str "true") (list prop #t))
         ((string-ci=? str "false") (list prop #f))
         (else #f)
       )
      )
      ((selection)
       (let*
         ((selected (send hierarchy-widget get-selected))
          (type     (send hierarchy-widget item-property-get selected 'type #f))
          (choices  (send hierarchy-widget item-property-get selected 'choices #f))
          (nmax     (length choices))
         )
         (cond
           ((member type '(choice radio-box))
            (if (and (integer? (string->number str)) 
                     (<= 0 (string->number str) (- nmax 1))
                )
              (list prop (string->number str))
              #f
            )
           )
           ((equal? type 'list-box)
            (cond
              ((and (integer? (string->number str))
                    (<= 0 (string->number str) (- nmax 1))
               )
               (list prop (string->number str))
              )
              ; looks like #f is allowed only for initialization
;              ((string=? str "") 
;               (list prop #f))
              (else #f)
            )
           )
           (else
            (error (string-append "valid-property?: '"
                                  (symbol->string type)
                                  "' widget has unhandled 'selection' property"
                   ))
            #f
           )
         )
       )
      )
      ;; This property can be either a number or a string depending on the current widget.
      ((init-value)
       (let*
         ((selected (send hierarchy-widget get-selected))
          (type     (send hierarchy-widget item-property-get selected 'type #f)))
         (cond
           ((member type '(text-field combo-field)) 
            (list prop str))
           ((eq? type 'slider)
            (if (and (integer? (string->number str)) (<= -10000 (string->number str) 10000))
              (list prop (string->number str))
              #f
            )
           )
           (else
            (error (string-append "valid-property?: '"
                                  (symbol->string type)
                                  "' widget has unhandled 'init-value' property"))
            #f
           )
         )
       )
      )
      ; ------------------------------------------------------------------------
      ; specially handled properties,
      ; these properties have been split into components
      
      ;; alignment
      ((halignment)
       (let*
         ((selected (send hierarchy-widget get-selected))
          (align    (send hierarchy-widget item-property-get selected 'alignment #f)))
         (list 'alignment (cons (string->symbol str) (cdr align)))
       )
      )
      ((valignment)
       (let*
         ((selected (send hierarchy-widget get-selected))
          (align    (send hierarchy-widget item-property-get selected 'alignment #f)))
         (list 'alignment (cons (car align) (list (string->symbol str))))
       )
      )
      ; style options
      ((style-deleted)
       (style-prop-make 'deleted)
      )
      ((style-border)
       (let*
         ((selected   (send hierarchy-widget get-selected))
          (type       (send hierarchy-widget item-property-get selected 'type #f))
         )
         (if (equal? type 'tab-panel) ; tab-panel has a no-border style
           (style-no-prop-make 'no-border)
           (style-prop-make 'border)
         )
       )
      )
      ((style-no-resize-border)
       (style-no-prop-make 'no-resize-border)
      )
      ((style-no-caption)
       (style-no-prop-make 'no-caption)
      )
      ((style-no-system-menu)
       (style-no-prop-make 'no-system-menu)
      )
      ((style-toolbar-button)
       (style-prop-make 'toolbar-button)
      )
      ((style-hide-menu-bar)
       (style-prop-make 'hide-menu-bar)
      )
      ((style-float)
       (style-prop-make 'float)
      )
      ((style-metal)
       (style-prop-make 'metal)
      )
      ((style-direction)
       (let*
         ((selected  (send hierarchy-widget get-selected))
          (style     (send hierarchy-widget item-property-get selected 'style #f))
          (new-style (remove-list '(horizontal vertical) style))
         )
         (list 'style (cons (string->symbol str) new-style))
       )
      )
      ((style-label-direction)
       (let*
         ((selected  (send hierarchy-widget get-selected))
          (style     (send hierarchy-widget item-property-get selected 'style #f))
          (new-style (remove-list '(horizontal-label vertical-label) style))
         )
         (list 'style (cons (string->symbol str) new-style))
       )
      )
      ((style-plain)
       (style-prop-make 'plain)
      )
      ((style-hscroll)
       (style-prop-make 'hscroll)
      )
      ((style-vscroll)
       (style-prop-make 'vscroll)
      )
      ((style-password)
       (style-prop-make 'password)
      )
      ((style-control-border)
       (style-prop-make 'control-border)
      )
      ((style-resize-corner)
       (style-prop-make 'resize-corner)
      )
      ((style-gl)
       (style-prop-make 'gl)
      )
      ((style-no-autoclear)
       (style-no-prop-make 'no-autoclear)
      )
      ((style-transparent)
       (style-prop-make 'transparent)
      )
      ((style-selection)
       (let*
         ((selected  (send hierarchy-widget get-selected))
          (style     (send hierarchy-widget item-property-get selected 'style #f))
          (new-style (remove-list '(single multiple extended) style))
         )
         (list 'style (cons (string->symbol str) new-style))
       )
      )
      ((style-text-field)
       (let*
         ((selected  (send hierarchy-widget get-selected))
          (style     (send hierarchy-widget item-property-get selected 'style #f))
          (new-style (remove-list '(single multiple) style))
         )
         (list 'style (cons (string->symbol str) new-style))
       )
      )
      ((choices)
       (list 'choices str)
      )
      ; callbackand paint-callback are handled separately
      
      ; anything else is not handled
      (else 
       (error (string-append "valid-property?: property '"
                             (symbol->string prop)
                             "' is unhandled"
              ))
       #f
      )))
    
  
  
  ; this is the callback function when a value has changed
  (provide properties-func)
  (define (properties-func prop val)
    ; change of id is handled specially
    (if (equal? prop 'id)
      (let 
        ((exists?  (send hierarchy-widget item-exists? val))
         (selected (send hierarchy-widget get-selected))
        )
        (if exists?
          (begin
            (message-box "Error" 
                         (string-append "The id '"
                                        val
                                        "' already exists!")
                         properties-frame '(ok stop))
            #f
          )
          (begin
            (send hierarchy-widget item-id-set! selected val)
            #t
          )
        )
      )
      ; all other properties
      (let
        ((res (validate-property prop val)))
        (if res
          (let
            ((name (car res))
             (val  (cadr res))
             (selected (send hierarchy-widget get-selected))
            )
            ; futher special cases with dependence on each other
            (cond
              ; when choices changes, we have to reset the selection as well
              ((equal? name 'choices)
               (send hierarchy-widget item-property-set! selected 'selection 0 #t)
               (send properties-panel set-value 'selection 0)
              )
              ((equal? name 'min-value)
               (let*
                 ((init-val (string->number (send properties-panel get-text 'init-value))))
                 (if (not (and init-val (> init-val val)))
                   (begin
                     (send hierarchy-widget item-property-set! selected 'init-value val #t)
                     (send properties-panel set-value 'init-value val)))))
              ((equal? name 'max-value)
               (let*
                 ((init-val (string->number (send properties-panel get-text 'init-value))))
                 (if (not (and init-val (< init-val val)))
                   (begin
                     (send hierarchy-widget item-property-set! selected 'init-value val #t)
                     (send properties-panel set-value 'init-value val)))))
            )
            (send hierarchy-widget item-property-set! selected name val #t)
            #t
          )
          (begin
            (show-invalid-dialog-box val (symbol->string prop))
            #f
          )
        )
      )
    )
  )
  
  
  
) ; end of module

